/*
 * SingleInitialState.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: 20. 9. 2014
 *	  Author: Tomas Pecka
 */

#ifndef _SINGLE_INITIAL_STATE_H_
#define _SINGLE_INITIAL_STATE_H_

#include <alib/algorithm>
#include <alib/set>

#include <automaton/FSM/DFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/MultiInitialStateEpsilonNFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/CompactNFA.h>

#include <label/InitialStateLabel.h>
#include <common/createUnique.hpp>

namespace automaton {

namespace simplify {

/**
 * Algorithm for the conversion of multi-initial state finite automata to single-initial state finite automata.
 * This algorithm is an implementation of Melichar: Jazyky a překlady 2.46.
 *
 * @sa automaton::simplify::SingleInitialStateEpsilonTransition
 */
class SingleInitialState {
public:
	/**
	 * Converts multi-initial state automaton to a single-initial state automaton.
	 * @tparam SymbolType Type for input symbols.
	 * @tparam StateType Type for states.
	 * @param automaton automaton to convert
	 * @return an automaton equivalent to @p with only one initial state
	 */
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > convert(const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::EpsilonNFA < SymbolType, StateType > convert(const automaton::MultiInitialStateEpsilonNFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, StateType > convert(const automaton::DFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > convert(const automaton::NFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::EpsilonNFA < SymbolType, StateType > convert(const automaton::EpsilonNFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::ExtendedNFA < SymbolType, StateType > convert(const automaton::ExtendedNFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::CompactNFA < SymbolType, StateType > convert(const automaton::CompactNFA < SymbolType, StateType > & automaton);
};

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > SingleInitialState::convert(const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton) {
	// step 1, 3
	StateType q0 = common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), automaton.getStates ( ) );

	automaton::NFA < SymbolType, StateType > res ( q0 );
	res.setInputAlphabet ( automaton.getInputAlphabet ( ) );

	for( const StateType & q : automaton.getStates ( ) )
		res.addState( q );

	// step 2
	for ( const auto & q : automaton.getInitialStates ( ) )
		for(const auto & kv : automaton.getTransitionsFromState ( q ) )
			res.addTransition ( q0, kv.first.second, kv.second );

	for ( const auto & t : automaton.getTransitions ( ) )
		res.addTransition ( t.first.first, t.first.second, t.second );

	res.setFinalStates ( automaton.getFinalStates ( ) );

	// step 4, 5
	if ( ! ext::excludes ( automaton.getFinalStates ( ).begin ( ), automaton.getFinalStates ( ).end ( ), automaton.getInitialStates ( ).begin ( ), automaton.getInitialStates ( ).end ( ) ) )
		res.addFinalState ( q0 );

	return res;
}

template < class SymbolType, class StateType >
automaton::EpsilonNFA < SymbolType, StateType > SingleInitialState::convert(const automaton::MultiInitialStateEpsilonNFA < SymbolType, StateType > & automaton) {
	// step 1, 3
	StateType q0 = common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), automaton.getStates ( ) );

	automaton::EpsilonNFA < SymbolType, StateType > res ( q0 );
	res.setInputAlphabet ( automaton.getInputAlphabet ( ) );

	for( const StateType & q : automaton.getStates ( ) )
		res.addState( q );

	// step 2
	for ( const auto & q : automaton.getInitialStates ( ) )
		for(const auto & kv : automaton.getTransitionsFromState ( q ) )
			res.addTransition ( q0, kv.first.second, kv.second );

	for ( const auto & t : automaton.getTransitions ( ) )
		res.addTransition ( t.first.first, t.first.second, t.second );

	res.setFinalStates ( automaton.getFinalStates ( ) );

	// step 4, 5
	if ( ! ext::excludes ( automaton.getFinalStates ( ).begin ( ), automaton.getFinalStates ( ).end ( ), automaton.getInitialStates ( ).begin ( ), automaton.getInitialStates ( ).end ( ) ) )
		res.addFinalState ( q0 );

	return res;
}

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, StateType > SingleInitialState::convert(const automaton::DFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::EpsilonNFA < SymbolType, StateType > SingleInitialState::convert(const automaton::EpsilonNFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > SingleInitialState::convert(const automaton::NFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::CompactNFA < SymbolType, StateType > SingleInitialState::convert(const automaton::CompactNFA < SymbolType, StateType > & automaton) {
	return automaton;
}

template < class SymbolType, class StateType >
automaton::ExtendedNFA < SymbolType, StateType > SingleInitialState::convert(const automaton::ExtendedNFA < SymbolType, StateType > & automaton) {
	return automaton;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* _SINGLE_INITIAL_STATE_H_ */
