/*
 * Total.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: 20. 9. 2014
 *	  Author: Tomas Pecka
 */

#ifndef TOTAL_H_
#define TOTAL_H_

#include <exception/CommonException.h>

#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>
#include <label/FailStateLabel.h>
#include <common/createUnique.hpp>

namespace automaton {

namespace simplify {

/**
 * Algorithm that makes finite automaton's transition function total (every state has a transition for every symbol).
 * Implementation of Melichar: Jazyky a překlady, 2.22.
 */
class Total {
public:
	/**
	 * Makes a finite automaton's transition function total.
	 *
	 * @tparam SymbolType Type for input symbols.
	 * @tparam StateType Type for states.
	 * @param automaton automaton to alter
	 * @return an automaton equivalent to @p automaton with total transition function
	 */
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > total(const automaton::NFA < SymbolType, StateType > & automaton);

	/**
	 * @overload
	 */
	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, StateType > total(const automaton::DFA < SymbolType, StateType > & automaton);
};

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > Total::total(const automaton::NFA < SymbolType, StateType > & automaton) {
	automaton::NFA < SymbolType, StateType > res(automaton);
	StateType nullState = common::createUnique(label::FailStateLabel::instance < StateType > ( ), automaton.getStates());
	res.addState(nullState);

	for(const auto& q : res.getStates()) {
		for(const auto& a : res.getInputAlphabet()) {
			if ( res.getTransitions ( ).find ( ext::make_pair ( q, a ) ) == res.getTransitions ( ).end ( ) ) {
				res.addTransition ( q, a, nullState );
			}
		}
	}

	return res;
}

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, StateType > Total::total(const automaton::DFA < SymbolType, StateType > & automaton) {
	automaton::DFA < SymbolType, StateType > res(automaton);
	StateType nullState = common::createUnique(label::FailStateLabel::instance < StateType > ( ), automaton.getStates());
	res.addState(nullState);

	for(const auto& q : res.getStates()) {
		for(const auto& a : res.getInputAlphabet()) {
			if(res.getTransitions().find(ext::make_pair(q, a)) == res.getTransitions().end()) {
				res.addTransition(q, a, nullState);
			}
		}
	}

	return res;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* TOTAL_H_ */
